@@ -15,6 +15,11 @@ over `Array{Bool, N}` and allowing some operations to work on 64 values at once.
 By default, Julia returns `BitArrays` from [broadcasting](@ref Broadcasting) operations
 that generate boolean elements (including dotted-comparisons like `.==`) as well as from
 the functions [`trues`](@ref) and [`falses`](@ref).
+
+!!! note
+    Due to its packed storage format, concurrent access to the elements of a `BitArray`
+    where at least one of them is a write is not thread safe.
+
 """
 mutable struct BitArray{N} <: AbstractArray{Bool, N}
     chunks::Vector{UInt64}
@@ -50,26 +55,48 @@ Behaves identically to the [`Array`](@ref) constructor. See [`undef`](@ref).
 # Examples
 ```julia-repl
 julia> BitArray(undef, 2, 2)
-2×2 BitArray{2}:
- false  false
- false  true
+2×2 BitMatrix:
+ 0  0
+ 0  0
 
 julia> BitArray(undef, (3, 1))
-3×1 BitArray{2}:
- false
- true
- false
+3×1 BitMatrix:
+ 0
+ 0
+ 0
 ```
 """
 BitArray(::UndefInitializer, dims::Integer...) = BitArray(undef, map(Int,dims))
-BitArray{N}(::UndefInitializer, dims::Integer...) where {N} = BitArray{N}(undef, map(Int,dims))
+BitArray{N}(::UndefInitializer, dims::Integer...) where {N} = BitArray{N}(undef, map(Int,dims))::BitArray{N}
 BitArray(::UndefInitializer, dims::NTuple{N,Integer}) where {N} = BitArray{N}(undef, map(Int, dims)...)
 BitArray{N}(::UndefInitializer, dims::NTuple{N,Integer}) where {N} = BitArray{N}(undef, map(Int, dims)...)
 
 const BitVector = BitArray{1}
 const BitMatrix = BitArray{2}
 
-BitVector() = BitArray{1}(undef, 0)
+BitVector() = BitVector(undef, 0)
+
+"""
+    BitVector(nt::Tuple{Vararg{Bool}})
+
+Construct a `BitVector` from a tuple of `Bool`.
+# Examples
+```julia-repl
+julia> nt = (true, false, true, false)
+(true, false, true, false)
+
+julia> BitVector(nt)
+4-element BitVector:
+ 1
+ 0
+ 1
+ 0
+```
+"""
+function BitVector(nt::Tuple{Vararg{Bool}})
+    bv = BitVector(undef, length(nt))
+    bv .= nt
+end
 
 ## utility functions ##
 
@@ -92,11 +119,11 @@ const _msk64 = ~UInt64(0)
 @inline _div64(l) = l >> 6
 @inline _mod64(l) = l & 63
 @inline _blsr(x)= x & (x-1) #zeros the last set bit. Has native instruction on many archs. needed in multidimensional.jl
-@inline _msk_end(l::Integer) = _msk64 >>> _mod64(-l)
+@inline _msk_end(l::Int) = _msk64 >>> _mod64(-l)
 @inline _msk_end(B::BitArray) = _msk_end(length(B))
 num_bit_chunks(n::Int) = _div64(n+63)
 
-@inline get_chunks_id(i::Integer) = _div64(Int(i)-1)+1, _mod64(Int(i)-1)
+@inline get_chunks_id(i::Int) = _div64(i-1)+1, _mod64(i-1)
 
 function glue_src_bitchunks(src::Vector{UInt64}, k::Int, ks1::Int, msk_s0::UInt64, ls0::Int)
     @inbounds begin
@@ -109,7 +136,7 @@ function glue_src_bitchunks(src::Vector{UInt64}, k::Int, ks1::Int, msk_s0::UInt6
     return chunk
 end
 
-function copy_chunks!(dest::Vector{UInt64}, pos_d::Integer, src::Vector{UInt64}, pos_s::Integer, numbits::Integer)
+function copy_chunks!(dest::Vector{UInt64}, pos_d::Int, src::Vector{UInt64}, pos_s::Int, numbits::Int)
     numbits == 0 && return
     if dest === src && pos_d > pos_s
         return copy_chunks_rtol!(dest, pos_d, pos_s, numbits)
@@ -165,7 +192,7 @@ function copy_chunks!(dest::Vector{UInt64}, pos_d::Integer, src::Vector{UInt64},
     return
 end
 
-function copy_chunks_rtol!(chunks::Vector{UInt64}, pos_d::Integer, pos_s::Integer, numbits::Integer)
+function copy_chunks_rtol!(chunks::Vector{UInt64}, pos_d::Int, pos_s::Int, numbits::Int)
     pos_d == pos_s && return
     pos_d < pos_s && return copy_chunks!(chunks, pos_d, chunks, pos_s, numbits)
 
@@ -213,7 +240,7 @@ function copy_chunks_rtol!(chunks::Vector{UInt64}, pos_d::Integer, pos_s::Intege
     end
 end
 
-function fill_chunks!(Bc::Array{UInt64}, x::Bool, pos::Integer, numbits::Integer)
+function fill_chunks!(Bc::Array{UInt64}, x::Bool, pos::Int, numbits::Int)
     numbits <= 0 && return
     k0, l0 = get_chunks_id(pos)
     k1, l1 = get_chunks_id(pos+numbits-1)
@@ -368,7 +395,7 @@ Create a `BitArray` with all values set to `false`.
 # Examples
 ```jldoctest
 julia> falses(2,3)
-2×3 BitArray{2}:
+2×3 BitMatrix:
  0  0  0
  0  0  0
 ```
@@ -386,7 +413,7 @@ Create a `BitArray` with all values set to `true`.
 # Examples
 ```jldoctest
 julia> trues(2,3)
-2×3 BitArray{2}:
+2×3 BitMatrix:
  1  1  1
  1  1  1
 ```
@@ -427,11 +454,13 @@ function copyto!(dest::BitArray, src::BitArray)
 end
 
 function unsafe_copyto!(dest::BitArray, doffs::Integer, src::Union{BitArray,Array}, soffs::Integer, n::Integer)
-    copy_to_bitarray_chunks!(dest.chunks, doffs, src, soffs, n)
+    copy_to_bitarray_chunks!(dest.chunks, Int(doffs), src, Int(soffs), Int(n))
     return dest
 end
 
-function copyto!(dest::BitArray, doffs::Integer, src::Array, soffs::Integer, n::Integer)
+copyto!(dest::BitArray, doffs::Integer, src::Array, soffs::Integer, n::Integer) =
+    _copyto_int!(dest, Int(doffs), src, Int(soffs), Int(n))
+function _copyto_int!(dest::BitArray, doffs::Int, src::Array, soffs::Int, n::Int)
     n == 0 && return dest
     soffs < 1 && throw(BoundsError(src, soffs))
     doffs < 1 && throw(BoundsError(dest, doffs))
@@ -473,7 +502,7 @@ end
 
 BitArray(A::AbstractArray{<:Any,N}) where {N} = BitArray{N}(A)
 function BitArray{N}(A::AbstractArray{T,N}) where N where T
-    B = BitArray(undef, size(A))
+    B = BitArray(undef, convert(Dims{N}, size(A)::Dims{N}))
     Bc = B.chunks
     l = length(B)
     l == 0 && return B
@@ -482,14 +511,14 @@ function BitArray{N}(A::AbstractArray{T,N}) where N where T
         for i = 1:length(Bc)-1
             c = UInt64(0)
             for j = 0:63
-                c |= (UInt64(convert(Bool, A[ind])) << j)
+                c |= (UInt64(convert(Bool, A[ind])::Bool) << j)
                 ind += 1
             end
             Bc[i] = c
         end
         c = UInt64(0)
         for j = 0:_mod64(l-1)
-            c |= (UInt64(convert(Bool, A[ind])) << j)
+            c |= (UInt64(convert(Bool, A[ind])::Bool) << j)
             ind += 1
         end
         Bc[end] = c
@@ -503,14 +532,14 @@ function BitArray{N}(A::Array{Bool,N}) where N
     l = length(B)
     l == 0 && return B
     copy_to_bitarray_chunks!(Bc, 1, A, 1, l)
-    return B
+    return B::BitArray{N}
 end
 
 reinterpret(::Type{Bool}, B::BitArray, dims::NTuple{N,Int}) where {N} = reinterpret(B, dims)
 reinterpret(B::BitArray, dims::NTuple{N,Int}) where {N} = reshape(B, dims)
 
 if nameof(@__MODULE__) === :Base  # avoid method overwrite
-(::Type{T})(x::T) where {T<:BitArray} = copy(x)
+(::Type{T})(x::T) where {T<:BitArray} = copy(x)::T
 BitArray(x::BitArray) = copy(x)
 end
 
@@ -523,17 +552,17 @@ The shape is inferred from the `itr` object.
 # Examples
 ```jldoctest
 julia> BitArray([1 0; 0 1])
-2×2 BitArray{2}:
+2×2 BitMatrix:
  1  0
  0  1
 
 julia> BitArray(x+y == 3 for x = 1:2, y = 1:3)
-2×3 BitArray{2}:
+2×3 BitMatrix:
  0  1  0
  1  0  0
 
 julia> BitArray(x+y == 3 for x = 1:2 for y = 1:3)
-6-element BitArray{1}:
+6-element BitVector:
  0
  1
  0
@@ -543,6 +572,7 @@ julia> BitArray(x+y == 3 for x = 1:2 for y = 1:3)
 ```
 """
 BitArray(itr) = gen_bitarray(IteratorSize(itr), itr)
+BitArray{N}(itr) where N = gen_bitarrayN(BitArray{N}, IteratorSize(itr), itr)
 
 convert(T::Type{<:BitArray}, a::AbstractArray) = a isa T ? a : T(a)
 
@@ -571,6 +601,14 @@ end
 
 gen_bitarray(::IsInfinite, itr) =  throw(ArgumentError("infinite-size iterable used in BitArray constructor"))
 
+gen_bitarrayN(::Type{BitVector}, itsz, itr)                        = gen_bitarray(itsz, itr)
+gen_bitarrayN(::Type{BitVector}, itsz::HasShape{1}, itr)           = gen_bitarray(itsz, itr)
+gen_bitarrayN(::Type{BitArray{N}}, itsz::HasShape{N}, itr) where N = gen_bitarray(itsz, itr)
+# The first of these is just for ambiguity resolution
+gen_bitarrayN(::Type{BitVector}, itsz::HasShape{N}, itr) where N      = throw(DimensionMismatch("cannot create a BitVector from a $N-dimensional iterator"))
+gen_bitarrayN(@nospecialize(T::Type), itsz::HasShape{N}, itr) where N = throw(DimensionMismatch("cannot create a $T from a $N-dimensional iterator"))
+gen_bitarrayN(@nospecialize(T::Type), itsz, itr) = throw(DimensionMismatch("cannot create a $T from a generic iterator"))
+
 # The aux functions gen_bitarray_from_itr and fill_bitarray_from_itr! both
 # use a Vector{Bool} cache for performance reasons
 
@@ -741,7 +779,7 @@ function append!(B::BitVector, items::BitVector)
     return B
 end
 
-append!(B::BitVector, items) = append!(B, BitArray(items))
+append!(B::BitVector, items) = append!(B, BitVector(items))
 append!(A::Vector{Bool}, items::BitVector) = append!(A, Array(items))
 
 function prepend!(B::BitVector, items::BitVector)
@@ -769,7 +807,8 @@ function sizehint!(B::BitVector, sz::Integer)
     return B
 end
 
-function resize!(B::BitVector, n::Integer)
+resize!(B::BitVector, n::Integer) = _resize_int!(B, Int(n))
+function _resize_int!(B::BitVector, n::Int)
     n0 = length(B)
     n == n0 && return B
     n >= 0 || throw(BoundsError(B, n))
@@ -779,7 +818,7 @@ function resize!(B::BitVector, n::Integer)
     end
     Bc = B.chunks
     k0 = length(Bc)
-    k1 = num_bit_chunks(Int(n))
+    k1 = num_bit_chunks(n)
     if k1 > k0
         _growend!(Bc, k1 - k0)
         Bc[end] = UInt64(0)
@@ -845,7 +884,9 @@ function popfirst!(B::BitVector)
     return item
 end
 
-function insert!(B::BitVector, i::Integer, item)
+insert!(B::BitVector, i::Integer, item) = _insert_int!(B, Int(i), item)
+function _insert_int!(B::BitVector, i::Int, item)
+    i = Int(i)
     n = length(B)
     1 <= i <= n+1 || throw(BoundsError(B, i))
     item = convert(Bool, item)
@@ -872,7 +913,7 @@ function insert!(B::BitVector, i::Integer, item)
     B
 end
 
-function _deleteat!(B::BitVector, i::Integer)
+function _deleteat!(B::BitVector, i::Int)
     k, j = get_chunks_id(i)
 
     msk_bef = _msk64 >>> (63 - j)
@@ -906,6 +947,8 @@ function _deleteat!(B::BitVector, i::Integer)
 end
 
 function deleteat!(B::BitVector, i::Integer)
+    i isa Bool && depwarn("passing Bool as an index is deprecated", :deleteat!)
+    i = Int(i)
     n = length(B)
     1 <= i <= n || throw(BoundsError(B, i))
 
@@ -940,30 +983,32 @@ function deleteat!(B::BitVector, inds)
     n = new_l = length(B)
     y = iterate(inds)
     y === nothing && return B
-    n == 0 && throw(BoundsError(B, inds))
 
     Bc = B.chunks
 
     (p, s) = y
+    checkbounds(B, p)
+    p isa Bool && throw(ArgumentError("invalid index $p of type Bool"))
     q = p+1
     new_l -= 1
     y = iterate(inds, s)
     while y !== nothing
         (i, s) = y
         if !(q <= i <= n)
+            i isa Bool && throw(ArgumentError("invalid index $i of type Bool"))
             i < q && throw(ArgumentError("indices must be unique and sorted"))
             throw(BoundsError(B, i))
         end
         new_l -= 1
         if i > q
-            copy_chunks!(Bc, p, Bc, q, i-q)
+            copy_chunks!(Bc, Int(p), Bc, Int(q), Int(i-q))
             p += i-q
         end
         q = i+1
         y = iterate(inds, s)
     end
 
-    q <= n && copy_chunks!(Bc, p, Bc, q, n-q+1)
+    q <= n && copy_chunks!(Bc, Int(p), Bc, Int(q), Int(n-q+1))
 
     delta_k = num_bit_chunks(new_l) - length(Bc)
     delta_k < 0 && _deleteend!(Bc, -delta_k)
@@ -977,7 +1022,56 @@ function deleteat!(B::BitVector, inds)
     return B
 end
 
+function deleteat!(B::BitVector, inds::AbstractVector{Bool})
+    length(inds) == length(B) || throw(BoundsError(B, inds))
+
+    n = new_l = length(B)
+    y = findfirst(inds)
+    y === nothing && return B
+
+    Bc = B.chunks
+
+    p = y
+    s = y + 1
+    checkbounds(B, p)
+    q = p + 1
+    new_l -= 1
+    y = findnext(inds, s)
+    while y !== nothing
+        i = y
+        s = y + 1
+        new_l -= 1
+        if i > q
+            copy_chunks!(Bc, Int(p), Bc, Int(q), Int(i-q))
+            p += i - q
+        end
+        q = i + 1
+        y = findnext(inds, s)
+    end
+
+    q <= n && copy_chunks!(Bc, Int(p), Bc, Int(q), Int(n - q + 1))
+
+    delta_k = num_bit_chunks(new_l) - length(Bc)
+    delta_k < 0 && _deleteend!(Bc, -delta_k)
+
+    B.len = new_l
+
+    if new_l > 0
+        Bc[end] &= _msk_end(new_l)
+    end
+
+    return B
+end
+
+keepat!(B::BitVector, inds) = _keepat!(B, inds)
+keepat!(B::BitVector, inds::AbstractVector{Bool}) = _keepat!(B, inds)
+
 function splice!(B::BitVector, i::Integer)
+    # TODO: after deprecation remove the four lines below
+    #       as v = B[i] is enough to do both bounds checking
+    #       and Bool check then just pass Int(i) to _deleteat!
+    i isa Bool && depwarn("passing Bool as an index is deprecated", :splice!)
+    i = Int(i)
     n = length(B)
     1 <= i <= n || throw(BoundsError(B, i))
 
@@ -989,10 +1083,13 @@ end
 const _default_bit_splice = BitVector()
 
 function splice!(B::BitVector, r::Union{UnitRange{Int}, Integer}, ins::AbstractArray = _default_bit_splice)
-    n = length(B)
-    i_f = first(r)
-    i_l = last(r)
+    r isa Bool && depwarn("passing Bool as an index is deprecated", :splice!)
+    _splice_int!(B, isa(r, UnitRange{Int}) ? r : Int(r), ins)
+end
 
+function _splice_int!(B::BitVector, r, ins)
+    n = length(B)
+    i_f, i_l = first(r), last(r)
     1 <= i_f <= n+1 || throw(BoundsError(B, i_f))
     i_l <= n || throw(BoundsError(B, n+1))
 
@@ -1119,8 +1216,9 @@ end
 
 # TODO some of this could be optimized
 
-function reverse(A::BitArray; dims::Integer)
-    nd = ndims(A); d = dims
+_reverse(A::BitArray, d::Tuple{Integer}) = _reverse(A, d[1])
+function _reverse(A::BitArray, d::Int)
+    nd = ndims(A)
     1 ≤ d ≤ nd || throw(ArgumentError("dimension $d is not 1 ≤ $d ≤ $nd"))
     sd = size(A, d)
     sd == 1 && return copy(A)
@@ -1129,7 +1227,7 @@ function reverse(A::BitArray; dims::Integer)
 
     nnd = 0
     for i = 1:nd
-        nnd += Int(size(A,i)==1 || i==d)
+        nnd += size(A,i)==1 || i==d
     end
     if nnd == nd
         # reverse along the only non-singleton dimension
@@ -1165,17 +1263,7 @@ function reverse(A::BitArray; dims::Integer)
     return B
 end
 
-function reverse_bits(src::UInt64)
-    z    = src
-    z    = ((z >>>  1) & 0x5555555555555555) | ((z <<  1) & 0xaaaaaaaaaaaaaaaa)
-    z    = ((z >>>  2) & 0x3333333333333333) | ((z <<  2) & 0xcccccccccccccccc)
-    z    = ((z >>>  4) & 0x0f0f0f0f0f0f0f0f) | ((z <<  4) & 0xf0f0f0f0f0f0f0f0)
-    z    = ((z >>>  8) & 0x00ff00ff00ff00ff) | ((z <<  8) & 0xff00ff00ff00ff00)
-    z    = ((z >>> 16) & 0x0000ffff0000ffff) | ((z << 16) & 0xffff0000ffff0000)
-    return ((z >>> 32) & 0x00000000ffffffff) | ((z << 32) & 0xffffffff00000000)
-end
-
-function reverse!(B::BitVector)
+function _reverse!(B::BitVector, ::Colon)
     # Basic idea: each chunk is divided into two blocks of size k = n % 64, and
     # h = 64 - k. Walk from either end (with indices i and j) reversing chunks
     # and separately ORing their two blocks into place.
@@ -1198,14 +1286,14 @@ function reverse!(B::BitVector)
 
     i, j = 0, length(B.chunks)
     u = UInt64(0)
-    v = reverse_bits(B.chunks[j])
+    v = bitreverse(B.chunks[j])
     B.chunks[j] = 0
     @inbounds while true
         i += 1
         if i == j
             break
         end
-        u = reverse_bits(B.chunks[i])
+        u = bitreverse(B.chunks[i])
         B.chunks[i] = 0
         B.chunks[j] |= u >>> h
         B.chunks[i] |= v >>> h
@@ -1214,7 +1302,7 @@ function reverse!(B::BitVector)
         if i == j
             break
         end
-        v = reverse_bits(B.chunks[j])
+        v = bitreverse(B.chunks[j])
         B.chunks[j] = 0
         B.chunks[i] |= v << k
         B.chunks[j] |= u << k
@@ -1229,14 +1317,11 @@ function reverse!(B::BitVector)
     return B
 end
 
-reverse(v::BitVector) = reverse!(copy(v))
-
-
 function (<<)(B::BitVector, i::UInt)
     n = length(B)
     i == 0 && return copy(B)
     A = falses(n)
-    i < n && copy_chunks!(A.chunks, 1, B.chunks, i+1, n-i)
+    i < n && copy_chunks!(A.chunks, 1, B.chunks, Int(i+1), Int(n-i))
     return A
 end
 
@@ -1244,7 +1329,7 @@ function (>>>)(B::BitVector, i::UInt)
     n = length(B)
     i == 0 && return copy(B)
     A = falses(n)
-    i < n && copy_chunks!(A.chunks, i+1, B.chunks, 1, n-i)
+    i < n && copy_chunks!(A.chunks, Int(i+1), B.chunks, 1, Int(n-i))
     return A
 end
 
@@ -1259,7 +1344,7 @@ values. If `n < 0`, elements are shifted backwards. Equivalent to
 # Examples
 ```jldoctest
 julia> B = BitVector([true, false, true, false, false])
-5-element BitArray{1}:
+5-element BitVector:
  1
  0
  1
@@ -1267,7 +1352,7 @@ julia> B = BitVector([true, false, true, false, false])
  0
 
 julia> B >> 1
-5-element BitArray{1}:
+5-element BitVector:
  0
  1
  0
@@ -1275,7 +1360,7 @@ julia> B >> 1
  0
 
 julia> B >> -1
-5-element BitArray{1}:
+5-element BitVector:
  0
  1
  0
@@ -1297,7 +1382,7 @@ values. If `n < 0`, elements are shifted forwards. Equivalent to
 # Examples
 ```jldoctest
 julia> B = BitVector([true, false, true, false, false])
-5-element BitArray{1}:
+5-element BitVector:
  1
  0
  1
@@ -1305,7 +1390,7 @@ julia> B = BitVector([true, false, true, false, false])
  0
 
 julia> B << 1
-5-element BitArray{1}:
+5-element BitVector:
  0
  1
  0
@@ -1313,7 +1398,7 @@ julia> B << 1
  0
 
 julia> B << -1
-5-element BitArray{1}:
+5-element BitVector:
  0
  1
  0
@@ -1331,7 +1416,9 @@ details and examples.
 """
 (>>>)(B::BitVector, i::Int) = (i >=0 ? B >> unsigned(i) : B << unsigned(-i))
 
-function circshift!(dest::BitVector, src::BitVector, i::Integer)
+circshift!(dest::BitVector, src::BitVector, i::Integer) = _circshift_int!(dest, src, Int(i))
+function _circshift_int!(dest::BitVector, src::BitVector, i::Int)
+    i = Int(i)
     length(dest) == length(src) || throw(ArgumentError("destination and source should be of same size"))
     n = length(dest)
     i %= n
@@ -1352,17 +1439,17 @@ circshift!(B::BitVector, i::Integer) = circshift!(B, B, i)
 
 ## count & find ##
 
-function bitcount(Bc::Vector{UInt64})
-    n = 0
+function bitcount(Bc::Vector{UInt64}; init::T=0) where {T}
+    n::T = init
     @inbounds for i = 1:length(Bc)
-        n += count_ones(Bc[i])
+        n = (n + count_ones(Bc[i])) % T
     end
     return n
 end
 
-count(B::BitArray) = bitcount(B.chunks)
+_count(::typeof(identity), B::BitArray, ::Colon, init) = bitcount(B.chunks; init)
 
-function unsafe_bitfindnext(Bc::Vector{UInt64}, start::Integer)
+function unsafe_bitfindnext(Bc::Vector{UInt64}, start::Int)
     chunk_start = _div64(start-1)+1
     within_chunk_start = _mod64(start-1)
     mask = _msk64 << within_chunk_start
@@ -1383,6 +1470,7 @@ end
 
 # returns the index of the next true element, or nothing if all false
 function findnext(B::BitArray, start::Integer)
+    start = Int(start)
     start > 0 || throw(BoundsError(B, start))
     start > length(B) && return nothing
     unsafe_bitfindnext(B.chunks, start)
@@ -1391,7 +1479,7 @@ end
 #findfirst(B::BitArray) = findnext(B, 1)  ## defined in array.jl
 
 # aux function: same as findnext(~B, start), but performed without temporaries
-function findnextnot(B::BitArray, start::Integer)
+function findnextnot(B::BitArray, start::Int)
     start > 0 || throw(BoundsError(B, start))
     start > length(B) && return nothing
 
@@ -1426,14 +1514,15 @@ findfirstnot(B::BitArray) = findnextnot(B,1)
 function findnext(pred::Fix2{<:Union{typeof(isequal),typeof(==)},Bool},
                   B::BitArray, start::Integer)
     v = pred.x
-    v == false && return findnextnot(B, start)
+    v == false && return findnextnot(B, Int(start))
     v == true && return findnext(B, start)
     return nothing
 end
 #findfirst(B::BitArray, v) = findnext(B, 1, v)  ## defined in array.jl
 
 # returns the index of the first element for which the function returns true
-function findnext(testf::Function, B::BitArray, start::Integer)
+findnext(testf::Function, B::BitArray, start::Integer) = _findnext_int(testf, B, Int(start))
+function _findnext_int(testf::Function, B::BitArray, start::Int)
     f0::Bool = testf(false)
     f1::Bool = testf(true)
     !f0 && f1 && return findnext(B, start)
@@ -1441,12 +1530,12 @@ function findnext(testf::Function, B::BitArray, start::Integer)
 
     start > 0 || throw(BoundsError(B, start))
     start > length(B) && return nothing
-    f0 && f1 && return Int(start)
+    f0 && f1 && return start
     return nothing # last case: !f0 && !f1
 end
 #findfirst(testf::Function, B::BitArray) = findnext(testf, B, 1)  ## defined in array.jl
 
-function unsafe_bitfindprev(Bc::Vector{UInt64}, start::Integer)
+function unsafe_bitfindprev(Bc::Vector{UInt64}, start::Int)
     chunk_start = _div64(start-1)+1
     mask = _msk_end(start)
 
@@ -1466,12 +1555,14 @@ end
 
 # returns the index of the previous true element, or nothing if all false
 function findprev(B::BitArray, start::Integer)
+    start = Int(start)
     start > 0 || return nothing
     start > length(B) && throw(BoundsError(B, start))
     unsafe_bitfindprev(B.chunks, start)
 end
 
-function findprevnot(B::BitArray, start::Integer)
+function findprevnot(B::BitArray, start::Int)
+    start = Int(start)
     start > 0 || return nothing
     start > length(B) && throw(BoundsError(B, start))
 
@@ -1499,14 +1590,15 @@ findlastnot(B::BitArray) = findprevnot(B, length(B))
 function findprev(pred::Fix2{<:Union{typeof(isequal),typeof(==)},Bool},
                   B::BitArray, start::Integer)
     v = pred.x
-    v == false && return findprevnot(B, start)
+    v == false && return findprevnot(B, Int(start))
     v == true && return findprev(B, start)
     return nothing
 end
 #findlast(B::BitArray, v) = findprev(B, 1, v)  ## defined in array.jl
 
 # returns the index of the previous element for which the function returns true
-function findprev(testf::Function, B::BitArray, start::Integer)
+findprev(testf::Function, B::BitArray, start::Integer) = _findprev_int(testf, B, Int(start))
+function _findprev_int(testf::Function, B::BitArray, start::Int)
     f0::Bool = testf(false)
     f1::Bool = testf(true)
     !f0 && f1 && return findprev(B, start)
@@ -1514,7 +1606,7 @@ function findprev(testf::Function, B::BitArray, start::Integer)
 
     start > 0 || return nothing
     start > length(B) && throw(BoundsError(B, start))
-    f0 && f1 && return Int(start)
+    f0 && f1 && return start
     return nothing # last case: !f0 && !f1
 end
 #findlast(testf::Function, B::BitArray) = findprev(testf, B, 1)  ## defined in array.jl
@@ -1668,6 +1760,8 @@ map!(::typeof(identity), dest::BitArray, A::BitArray) = copyto!(dest, A)
 for (T, f) in ((:(Union{typeof(&), typeof(*), typeof(min)}), :(&)),
                (:(Union{typeof(|), typeof(max)}),            :(|)),
                (:(Union{typeof(xor), typeof(!=)}),           :xor),
+               (:(typeof(nand)),                             :nand),
+               (:(typeof(nor)),                              :nor),
                (:(Union{typeof(>=), typeof(^)}),             :((p, q) -> p | ~q)),
                (:(typeof(<=)),                               :((p, q) -> ~p | q)),
                (:(typeof(==)),                               :((p, q) -> ~xor(p, q))),
@@ -1768,7 +1862,12 @@ end
 
 function vcat(A::BitMatrix...)
     nargs = length(A)
-    nrows = sum(a->size(a, 1), A)::Int
+    nrows, nrowsA = 0, sizehint!(Int[], nargs)
+    for a in A
+        sz1 = size(a, 1)
+        nrows += sz1
+        push!(nrowsA, sz1)
+    end
     ncols = size(A[1], 2)
     for j = 2:nargs
         size(A[j], 2) == ncols ||
@@ -1776,12 +1875,10 @@ function vcat(A::BitMatrix...)
     end
     B = BitMatrix(undef, nrows, ncols)
     Bc = B.chunks
-    nrowsA = [size(a, 1) for a in A]
-    Ac = [a.chunks for a in A]
     pos_d = 1
     pos_s = fill(1, nargs)
     for j = 1:ncols, k = 1:nargs
-        copy_chunks!(Bc, pos_d, Ac[k], pos_s[k], nrowsA[k])
+        copy_chunks!(Bc, pos_d, A[k].chunks, pos_s[k], nrowsA[k])
         pos_s[k] += nrowsA[k]
         pos_d += nrowsA[k]
     end
@@ -1789,9 +1886,11 @@ function vcat(A::BitMatrix...)
 end
 
 # general case, specialized for BitArrays and Integers
-function _cat(dims::Integer, X::Union{BitArray, Bool}...)
+_cat(dims::Integer, X::Union{BitArray, Bool}...) = _cat(Int(dims)::Int, X...)
+function _cat(dims::Int, X::Union{BitArray, Bool}...)
+    dims = Int(dims)
     catdims = dims2cat(dims)
-    shape = cat_shape(catdims, (), map(cat_size, X)...)
+    shape = cat_shape(catdims, map(cat_size, X))
     A = falses(shape)
     return __cat(A, shape, catdims, X...)
 end
