@@ -33,17 +33,6 @@ function scan_entry!(result::Vector{SlotInfo}, idx::Int, @nospecialize(stmt))
 end
 
 
-function lift_defuse(cfg::CFG, defuse)
-    map(defuse) do slot
-        SlotInfo(
-            Int[block_for_inst(cfg, x) for x in slot.defs],
-            Int[block_for_inst(cfg, x) for x in slot.uses],
-            slot.any_newvar
-        )
-    end
-end
-
-@inline slot_id(s) = isa(s, SlotNumber) ? (s::SlotNumber).id : (s::TypedSlot).id
 function scan_slot_def_use(nargs::Int, ci::CodeInfo, code::Vector{Any})
     nslots = length(ci.slotflags)
     result = SlotInfo[SlotInfo() for i = 1:nslots]
@@ -51,27 +40,26 @@ function scan_slot_def_use(nargs::Int, ci::CodeInfo, code::Vector{Any})
     for var in result[1:(1+nargs)]
         push!(var.defs, 0)
     end
-    for (idx, stmt) in Iterators.enumerate(code)
+    for idx in 1:length(code)
+        stmt = code[idx]
         scan_entry!(result, idx, stmt)
     end
     result
 end
 
-function renumber_ssa(stmt::SSAValue, ssanums::Vector{Any}, new_ssa::Bool=false, used_ssa::Union{Nothing, Vector{Int}}=nothing)
+function renumber_ssa(stmt::SSAValue, ssanums::Vector{SSAValue}, new_ssa::Bool=false)
     id = stmt.id
     if id > length(ssanums)
         return stmt
     end
     val = ssanums[id]
-    if isa(val, SSAValue) && used_ssa !== nothing
-        used_ssa[val.id] += 1
-    end
+    @assert val.id > 0
     return val
 end
 
-function renumber_ssa!(@nospecialize(stmt), ssanums::Vector{Any}, new_ssa::Bool=false, used_ssa::Union{Nothing, Vector{Int}}=nothing)
-    isa(stmt, SSAValue) && return renumber_ssa(stmt, ssanums, new_ssa, used_ssa)
-    return ssamap(val->renumber_ssa(val, ssanums, new_ssa, used_ssa), stmt)
+function renumber_ssa!(@nospecialize(stmt), ssanums::Vector{SSAValue}, new_ssa::Bool=false)
+    isa(stmt, SSAValue) && return renumber_ssa(stmt, ssanums, new_ssa)
+    return ssamap(val->renumber_ssa(val, ssanums, new_ssa), stmt)
 end
 
 function make_ssa!(ci::CodeInfo, code::Vector{Any}, idx, slot, @nospecialize(typ))
@@ -83,10 +71,6 @@ function make_ssa!(ci::CodeInfo, code::Vector{Any}, idx, slot, @nospecialize(typ
     idx
 end
 
-struct UndefToken
-end
-const undef_token = UndefToken()
-
 function new_to_regular(@nospecialize(stmt), new_offset::Int)
     if isa(stmt, NewSSAValue)
         return SSAValue(stmt.id + new_offset)
@@ -105,17 +89,21 @@ function fixup_slot!(ir::IRCode, ci::CodeInfo, idx::Int, slot::Int, @nospecializ
     # We don't really have the information here to get rid of these.
     # We'll do so later
     if ssa === undef_token
-        insert_node!(ir, idx, Any, Expr(:throw_undef_if_not, ci.slotnames[slot], false))
+        insert_node!(ir, idx, NewInstruction(
+            Expr(:throw_undef_if_not, ci.slotnames[slot], false), Any))
         return undef_token
     end
     if !isa(ssa, Argument) && !(ssa === nothing) && ((ci.slotflags[slot] & SLOT_USEDUNDEF) != 0)
-        insert_node!(ir, idx, Any, Expr(:undefcheck, ci.slotnames[slot], ssa))
+        # insert a temporary node. type_lift_pass! will remove it
+        insert_node!(ir, idx, NewInstruction(
+            Expr(:undefcheck, ci.slotnames[slot], ssa), Any))
     end
     if isa(stmt, SlotNumber)
         return ssa
     elseif isa(stmt, TypedSlot)
-        return NewSSAValue(insert_node!(ir, idx, stmt.typ, PiNode(ssa, stmt.typ)).id - length(ir.stmts))
+        return NewSSAValue(insert_node!(ir, idx, NewInstruction(PiNode(ssa, stmt.typ), stmt.typ)).id - length(ir.stmts))
     end
+    @assert false # unreachable
 end
 
 function fixemup!(cond, rename, ir::IRCode, ci::CodeInfo, idx::Int, @nospecialize(stmt))
@@ -132,7 +120,7 @@ function fixemup!(cond, rename, ir::IRCode, ci::CodeInfo, idx::Int, @nospecializ
             val = stmt.values[i]
             isa(val, Union{SlotNumber, TypedSlot}) || continue
             cond(val) || continue
-            bb_idx = block_for_inst(ir.cfg, stmt.edges[i])
+            bb_idx = block_for_inst(ir.cfg, Int(stmt.edges[i]))
             from_bb_terminator = last(ir.cfg.blocks[bb_idx].stmts)
             stmt.values[i] = fixup_slot!(ir, ci, from_bb_terminator, slot_id(val), val, rename(val))
         end
@@ -152,6 +140,7 @@ function fixemup!(cond, rename, ir::IRCode, ci::CodeInfo, idx::Int, @nospecializ
                     return true
                 end
             end
+            # temporarily corrupt the isdefined node. type_lift_pass! will fix it
             stmt.args[1] = ssa
         end
         return stmt
@@ -167,12 +156,14 @@ function fixemup!(cond, rename, ir::IRCode, ci::CodeInfo, idx::Int, @nospecializ
                 return nothing
             end
             op[] = x
+        elseif isa(val, GlobalRef) && !(isdefined(val.mod, val.name) && isconst(val.mod, val.name))
+            op[] = NewSSAValue(insert_node!(ir, idx, NewInstruction(val, Any)).id - length(ir.stmts))
         end
     end
     return urs[]
 end
 
-function fixup_uses!(ir::IRCode, ci::CodeInfo, code, uses::Vector{Int}, slot, @nospecialize(ssa))
+function fixup_uses!(ir::IRCode, ci::CodeInfo, code::Vector{Any}, uses::Vector{Int}, slot::Int, @nospecialize(ssa))
     for use in uses
         code[use] = fixemup!(stmt->slot_id(stmt)==slot, stmt->ssa, ir, ci, use, code[use])
     end
@@ -182,7 +173,7 @@ function rename_uses!(ir::IRCode, ci::CodeInfo, idx::Int, @nospecialize(stmt), r
     return fixemup!(stmt->true, stmt->renames[slot_id(stmt)], ir, ci, idx, stmt)
 end
 
-function strip_trailing_junk!(ci::CodeInfo, code::Vector{Any}, flags::Vector{UInt8})
+function strip_trailing_junk!(ci::CodeInfo, code::Vector{Any}, info::Vector{Any}, flags::Vector{UInt8})
     # Remove `nothing`s at the end, we don't handle them well
     # (we expect the last instruction to be a terminator)
     for i = length(code):-1:1
@@ -190,6 +181,7 @@ function strip_trailing_junk!(ci::CodeInfo, code::Vector{Any}, flags::Vector{UIn
             resize!(code, i)
             resize!(ci.ssavaluetypes, i)
             resize!(ci.codelocs, i)
+            resize!(info, i)
             resize!(flags, i)
             break
         end
@@ -201,6 +193,7 @@ function strip_trailing_junk!(ci::CodeInfo, code::Vector{Any}, flags::Vector{UIn
         push!(code, ReturnNode())
         push!(ci.ssavaluetypes, Union{})
         push!(ci.codelocs, 0)
+        push!(info, nothing)
         push!(flags, 0x00)
     end
     nothing
@@ -352,8 +345,9 @@ end
 
 function rename_phinode_edges(node, bb, result_order, bb_rename)
     new_values = Any[]
-    new_edges = Any[]
+    new_edges = Int32[]
     for (idx, edge) in pairs(node.edges)
+        edge = Int(edge)
         (edge == 0 || haskey(bb_rename, edge)) || continue
         new_edge_from = edge == 0 ? 0 : rename_incoming_edge(edge, bb, result_order, bb_rename)
         push!(new_edges, new_edge_from)
@@ -383,7 +377,7 @@ function domsort_ssa!(ir::IRCode, domtree::DomTree)
     while node !== -1
         push!(result_order, node)
         cs = domtree.nodes[node].children
-        terminator = ir.stmts[last(ir.cfg.blocks[node].stmts)]
+        terminator = ir.stmts[last(ir.cfg.blocks[node].stmts)][:inst]
         iscondbr = isa(terminator, GotoIfNot)
         let old_node = node + 1
             if length(cs) >= 1
@@ -425,21 +419,23 @@ function domsort_ssa!(ir::IRCode, domtree::DomTree)
             nstmts += length(ir.cfg.blocks[i].stmts)
         end
     end
-    result_stmts = Vector{Any}(undef, nstmts + ncritbreaks + nnewfallthroughs)
-    result_types = Any[Any for i = 1:length(result_stmts)]
-    result_ltable = fill(Int32(0), length(result_stmts))
-    result_flags = fill(0x00, length(result_stmts))
-    inst_rename = Vector{Any}(undef, length(ir.stmts))
-    for i = 1:length(ir.new_nodes)
-        push!(inst_rename, SSAValue(nstmts + i + ncritbreaks + nnewfallthroughs))
+    result = InstructionStream(nstmts + ncritbreaks + nnewfallthroughs)
+    inst_rename = Vector{SSAValue}(undef, length(ir.stmts) + length(ir.new_nodes))
+    @inbounds for i = 1:length(ir.stmts)
+        inst_rename[i] = SSAValue(-1)
+    end
+    @inbounds for i = 1:length(ir.new_nodes)
+        inst_rename[i + length(ir.stmts)] = SSAValue(i + length(result))
     end
     bb_start_off = 0
     crit_edge_breaks_fixup = Tuple{Int, Int}[]
     for (new_bb, bb) in pairs(result_order)
         if bb == 0
-            @assert isa(result_stmts[bb_start_off+1], GotoNode)
+            nidx = bb_start_off + 1
+            inst = result[nidx][:inst]
+            @assert isa(inst, GotoNode)
             # N.B.: The .label has already been renamed when it was created.
-            new_bbs[new_bb] = BasicBlock((bb_start_off+1):(bb_start_off+1), [new_bb-1], [result_stmts[bb_start_off+1].label])
+            new_bbs[new_bb] = BasicBlock(nidx:nidx, [new_bb - 1], [inst.label])
             bb_start_off += 1
             continue
         end
@@ -447,41 +443,45 @@ function domsort_ssa!(ir::IRCode, domtree::DomTree)
         inst_range = (bb_start_off+1):(bb_start_off+length(old_inst_range))
         for (nidx, idx) in zip(inst_range, old_inst_range)
             inst_rename[idx] = SSAValue(nidx)
-            stmt = ir.stmts[idx]
-            if isa(stmt, PhiNode)
-                result_stmts[nidx] = rename_phinode_edges(stmt, bb, result_order, bb_rename)
-            else
-                result_stmts[nidx] = stmt
+            @assert !isassigned(result.inst, nidx)
+            node = result[nidx]
+            node[] = ir.stmts[idx]
+            inst = node[:inst]
+            if isa(inst, PhiNode)
+                node[:inst] = rename_phinode_edges(inst, bb, result_order, bb_rename)
             end
-            result_types[nidx] = ir.types[idx]
-            result_ltable[nidx] = ir.lines[idx]
-            result_flags[nidx] = ir.flags[idx]
         end
         # Now fix up the terminator
-        terminator = result_stmts[inst_range[end]]
+        terminator = result[inst_range[end]][:inst]
         if isa(terminator, GotoNode)
             # Convert to implicit fall through
             if bb_rename[terminator.label] == new_bb + 1
-                result_stmts[inst_range[end]] = nothing
+                result[inst_range[end]][:inst] = nothing
             else
-                result_stmts[inst_range[end]] = GotoNode(bb_rename[terminator.label])
+                result[inst_range[end]][:inst] = GotoNode(bb_rename[terminator.label])
             end
         elseif isa(terminator, GotoIfNot)
             # Check if we need to break the critical edge
             if bb_rename[bb + 1] != new_bb + 1
                 @assert result_order[new_bb + 1] == 0
                 # Add an explicit goto node in the next basic block (we accounted for this above)
-                result_stmts[inst_range[end]+1] = GotoNode(bb_rename[bb+1])
+                nidx = inst_range[end] + 1
+                node = result[nidx]
+                node[:inst], node[:type], node[:line] = GotoNode(bb_rename[bb + 1]), Any, 0
             end
-            result_stmts[inst_range[end]] = GotoIfNot(terminator.cond, bb_rename[terminator.dest])
+            result[inst_range[end]][:inst] = GotoIfNot(terminator.cond, bb_rename[terminator.dest])
         elseif !isa(terminator, ReturnNode)
-            if isa(terminator, Expr) && terminator.head == :enter
-                terminator.args[1] = bb_rename[terminator.args[1]]
+            if isa(terminator, Expr)
+                if terminator.head == :enter
+                    terminator.args[1] = bb_rename[terminator.args[1]]
+                end
             end
             if bb_rename[bb + 1] != new_bb + 1
                 # Add an explicit goto node
-                result_stmts[inst_range[end]+1] = GotoNode(bb_rename[bb+1])
-                inst_range = first(inst_range):(last(inst_range)+1)
+                nidx = inst_range[end] + 1
+                node = result[nidx]
+                node[:inst], node[:type], node[:line] = GotoNode(bb_rename[bb + 1]), Any, 0
+                inst_range = first(inst_range):(last(inst_range) + 1)
             end
         end
         bb_start_off += length(inst_range)
@@ -492,22 +492,28 @@ function domsort_ssa!(ir::IRCode, domtree::DomTree)
         end
         new_bbs[new_bb] = BasicBlock(inst_range, new_preds, new_succs)
     end
-    result_stmts = Any[renumber_ssa!(result_stmts[i], inst_rename, true) for i in 1:length(result_stmts)]
-    cfg = CFG(new_bbs, Int[first(bb.stmts) for bb in new_bbs[2:end]])
-    new_new_nodes = Vector{NewNode}(undef, length(ir.new_nodes))
-    for i = 1:length(ir.new_nodes)
-        entry = ir.new_nodes[i]
-        new_new_nodes[i] = NewNode(inst_rename[entry.pos].id, entry.attach_after, entry.typ,
-            renumber_ssa!(isa(entry.node, PhiNode) ?
-                rename_phinode_edges(entry.node, block_for_inst(ir.cfg, entry.pos), result_order, bb_rename) : entry.node,
-                inst_rename, true),
-            entry.line)
+    for i in 1:length(result)
+        result[i][:inst] = renumber_ssa!(result[i][:inst], inst_rename, true)
     end
-    new_ir = IRCode(ir, result_stmts, result_types, result_ltable, result_flags, cfg, new_new_nodes)
+    cfg = CFG(new_bbs, Int[first(bb.stmts) for bb in new_bbs[2:end]])
+    new_new_nodes = NewNodeStream(length(ir.new_nodes))
+    for i = 1:length(ir.new_nodes)
+        new_info = ir.new_nodes.info[i]
+        new_new_info = NewNodeInfo(inst_rename[new_info.pos].id, new_info.attach_after)
+        new_new_nodes.info[i] = new_new_info
+        new_node = new_new_nodes.stmts[i]
+        new_node[] = ir.new_nodes.stmts[i]
+        new_node_inst = new_node[:inst]
+        if isa(new_node_inst, PhiNode)
+            new_node_inst = rename_phinode_edges(new_node_inst, block_for_inst(ir.cfg, new_info.pos), result_order, bb_rename)
+        end
+        new_node[:inst] = renumber_ssa!(new_node_inst, inst_rename, true)
+    end
+    new_ir = IRCode(ir, result, cfg, new_new_nodes)
     return new_ir
 end
 
-function compute_live_ins(cfg::CFG, defuse)
+function compute_live_ins(cfg::CFG, defuse #=::Union{SlotInfo,SSADefUse}=#)
     # We remove from `uses` any block where all uses are dominated
     # by a def. This prevents insertion of dead phi nodes at the top
     # of such a block if that block happens to be in a loop
@@ -569,12 +575,14 @@ function recompute_type(node::Union{PhiNode, PhiCNode}, ci::CodeInfo, ir::IRCode
     return new_typ
 end
 
-function construct_ssa!(ci::CodeInfo, code::Vector{Any}, ir::IRCode, domtree::DomTree, defuse, nargs::Int, sptypes::Vector{Any},
-                        slottypes::Vector{Any})
+function construct_ssa!(ci::CodeInfo, ir::IRCode, domtree::DomTree,
+                        defuses::Vector{SlotInfo}, nargs::Int, slottypes::Vector{Any})
+    code = ir.stmts.inst
     cfg = ir.cfg
     left = Int[]
     catch_entry_blocks = Tuple{Int, Int}[]
-    for (idx, stmt) in pairs(code)
+    for idx in 1:length(code)
+        stmt = code[idx]
         if isexpr(stmt, :enter)
             push!(catch_entry_blocks, (block_for_inst(cfg, idx), block_for_inst(cfg, stmt.args[1])))
         end
@@ -597,7 +605,7 @@ function construct_ssa!(ci::CodeInfo, code::Vector{Any}, ir::IRCode, domtree::Do
     for (_, exc) in catch_entry_blocks
         phicnodes[exc] = Vector{Tuple{SlotNumber, NewSSAValue, PhiCNode}}()
     end
-    @timeit "idf" for (idx, slot) in Iterators.enumerate(defuse)
+    @timeit "idf" for (idx, slot) in Iterators.enumerate(defuses)
         # No uses => no need for phi nodes
         isempty(slot.uses) && continue
         # TODO: Restore this optimization
@@ -610,12 +618,13 @@ function construct_ssa!(ci::CodeInfo, code::Vector{Any}, ir::IRCode, domtree::Do
                 typ = MaybeUndef(Union{})
                 ssaval = nothing
                 for use in slot.uses[]
-                    insert_node!(ir, use, Union{}, Expr(:throw_undef_if_not, ci.slotnames[idx], false))
+                    insert_node!(ir, use,
+                        NewInstruction(Expr(:throw_undef_if_not, ci.slotnames[idx], false), Union{}))
                 end
                 fixup_uses!(ir, ci, code, slot.uses, idx, nothing)
             else
                 val = code[slot.defs[]].args[2]
-                typ = typ_for_val(val, ci, sptypes, slot.defs[], slottypes)
+                typ = typ_for_val(val, ci, ir.sptypes, slot.defs[], slottypes)
                 ssaval = SSAValue(make_ssa!(ci, code, slot.defs[], idx, typ))
                 fixup_uses!(ir, ci, code, slot.uses, idx, ssaval)
             end
@@ -629,7 +638,9 @@ function construct_ssa!(ci::CodeInfo, code::Vector{Any}, ir::IRCode, domtree::Do
                 # Create a PhiC node in the catch entry block and
                 # an upsilon node in the corresponding enter block
                 node = PhiCNode(Any[])
-                phic_ssa = NewSSAValue(insert_node!(ir, first_insert_for_bb(code, cfg, li), Union{}, node).id - length(ir.stmts))
+                phic_ssa = NewSSAValue(
+                    insert_node!(ir, first_insert_for_bb(code, cfg, li),
+                        NewInstruction(node, Union{})).id - length(ir.stmts))
                 push!(phicnodes[li], (SlotNumber(idx), phic_ssa, node))
                 # Inform IDF that we now have a def in the catch block
                 if !(li in live.def_bbs)
@@ -641,24 +652,26 @@ function construct_ssa!(ci::CodeInfo, code::Vector{Any}, ir::IRCode, domtree::Do
         for block in phiblocks
             push!(phi_slots[block], idx)
             node = PhiNode()
-            ssa = NewSSAValue(insert_node!(ir, first_insert_for_bb(code, cfg, block), Union{}, node).id - length(ir.stmts))
+            ssa = NewSSAValue(insert_node!(ir,
+                first_insert_for_bb(code, cfg, block), NewInstruction(node, Union{})).id - length(ir.stmts))
             push!(phi_nodes[block], ssa=>node)
         end
         push!(left, idx)
     end
     # Perform SSA renaming
     initial_incoming_vals = Any[
-        if 0 in defuse[x].defs
+        if 0 in defuses[x].defs
             Argument(x)
-        elseif !defuse[x].any_newvar
+        elseif !defuses[x].any_newvar
             undef_token
         else
-            SSAValue(-1)
+            SSAValue(-2)
         end for x in 1:length(ci.slotflags)
     ]
     worklist = Tuple{Int, Int, Vector{Any}}[(1, 0, initial_incoming_vals)]
     visited = BitSet()
     type_refine_phi = BitSet()
+    new_nodes = ir.new_nodes
     @timeit "SSA Rename" while !isempty(worklist)
         (item::Int, pred, incoming_vals) = pop!(worklist)
         # Rename existing phi nodes first, because their uses occur on the edge
@@ -670,10 +683,10 @@ function construct_ssa!(ci::CodeInfo, code::Vector{Any}, ir::IRCode, domtree::Do
             end
             isa(stmt, PhiNode) || continue
             for (edgeidx, edge) in pairs(stmt.edges)
-                from_bb = edge == 0 ? 0 : block_for_inst(cfg, edge)
+                from_bb = edge == 0 ? 0 : block_for_inst(cfg, Int(edge))
                 from_bb == pred || continue
                 isassigned(stmt.values, edgeidx) || break
-                stmt.values[edgeidx] = rename_uses!(ir, ci, edge, stmt.values[edgeidx], incoming_vals)
+                stmt.values[edgeidx] = rename_uses!(ir, ci, Int(edge), stmt.values[edgeidx], incoming_vals)
                 break
             end
         end
@@ -681,13 +694,13 @@ function construct_ssa!(ci::CodeInfo, code::Vector{Any}, ir::IRCode, domtree::Do
         for (idx, slot) in Iterators.enumerate(phi_slots[item])
             ssaval, node = phi_nodes[item][idx]
             incoming_val = incoming_vals[slot]
-            if incoming_val == SSAValue(-1)
+            if incoming_val === SSAValue(-1)
                 # Optimistically omit this path.
                 # Liveness analysis would probably have prevented us from inserting this phi node
                 continue
             end
             push!(node.edges, pred)
-            if incoming_val == undef_token
+            if incoming_val === undef_token
                 resize!(node.values, length(node.values)+1)
             else
                 push!(node.values, incoming_val)
@@ -697,13 +710,14 @@ function construct_ssa!(ci::CodeInfo, code::Vector{Any}, ir::IRCode, domtree::Do
             if isa(incoming_val, NewSSAValue)
                 push!(type_refine_phi, ssaval.id)
             end
-            typ = incoming_val == undef_token ? MaybeUndef(Union{}) : typ_for_val(incoming_val, ci, sptypes, -1, slottypes)
-            old_entry = ir.new_nodes[ssaval.id]
+            typ = incoming_val === undef_token ? MaybeUndef(Union{}) : typ_for_val(incoming_val, ci, ir.sptypes, -1, slottypes)
+            old_entry = new_nodes.stmts[ssaval.id]
             if isa(typ, DelayedTyp)
                 push!(type_refine_phi, ssaval.id)
             end
-            new_typ = isa(typ, DelayedTyp) ? Union{} : tmerge(old_entry.typ, typ)
-            ir.new_nodes[ssaval.id] = NewNode(old_entry.pos, old_entry.attach_after, new_typ, node, old_entry.line)
+            new_typ = isa(typ, DelayedTyp) ? Union{} : tmerge(old_entry[:type], typ)
+            old_entry[:type] = new_typ
+            old_entry[:inst] = node
             incoming_vals[slot] = ssaval
         end
         (item in visited) && continue
@@ -717,12 +731,13 @@ function construct_ssa!(ci::CodeInfo, code::Vector{Any}, ir::IRCode, domtree::Do
         eidx = findfirst(x->x[1] == item, catch_entry_blocks)
         if eidx !== nothing
             for (slot, _, node) in phicnodes[catch_entry_blocks[eidx][2]]
-                ivalundef = incoming_vals[slot_id(slot)] === undef_token
-                unode = ivalundef ? UpsilonNode() : UpsilonNode(incoming_vals[slot_id(slot)])
-                typ = ivalundef ? MaybeUndef(Union{}) : slottypes[slot_id(slot)]
+                ival = incoming_vals[slot_id(slot)]
+                ivalundef = ival === undef_token
+                unode = ivalundef ? UpsilonNode() : UpsilonNode(ival)
+                typ = ivalundef ? MaybeUndef(Union{}) : typ_for_val(ival, ci, ir.sptypes, -1, slottypes)
                 push!(node.values,
                     NewSSAValue(insert_node!(ir, first_insert_for_bb(code, cfg, item),
-                                 typ, unode, true).id - length(ir.stmts)))
+                                 NewInstruction(unode, typ), true).id - length(ir.stmts)))
             end
         end
         push!(visited, item)
@@ -743,11 +758,11 @@ function construct_ssa!(ci::CodeInfo, code::Vector{Any}, ir::IRCode, domtree::Do
                 if isexpr(stmt, :(=)) && isa(stmt.args[1], SlotNumber)
                     id = slot_id(stmt.args[1])
                     val = stmt.args[2]
-                    typ = typ_for_val(val, ci, sptypes, idx, slottypes)
+                    typ = typ_for_val(val, ci, ir.sptypes, idx, slottypes)
                     # Having undef_token appear on the RHS is possible if we're on a dead branch.
                     # Do something reasonable here, by marking the LHS as undef as well.
                     if val !== undef_token
-                        incoming_vals[id] = SSAValue(make_ssa!(ci, code, idx, id, typ))
+                        incoming_vals[id] = SSAValue(make_ssa!(ci, code, idx, id, typ)::Int)
                     else
                         code[idx] = nothing
                         incoming_vals[id] = undef_token
@@ -763,7 +778,7 @@ function construct_ssa!(ci::CodeInfo, code::Vector{Any}, ir::IRCode, domtree::Do
                                 typ = MaybeUndef(Union{})
                             end
                             push!(phicnodes[exc][cidx][3].values,
-                                NewSSAValue(insert_node!(ir, idx, typ, node, true).id - length(ir.stmts)))
+                                NewSSAValue(insert_node!(ir, idx, NewInstruction(node, typ), true).id - length(ir.stmts)))
                         end
                     end
                 end
@@ -784,9 +799,10 @@ function construct_ssa!(ci::CodeInfo, code::Vector{Any}, ir::IRCode, domtree::Do
         end
     end
     # Convert into IRCode form
-    new_code = ir.stmts
-    ssavalmap = Any[SSAValue(-1) for _ in 1:(length(ci.ssavaluetypes)+1)]
-    result_types = Any[Any for _ in 1:length(new_code)]
+    nstmts = length(ir.stmts)
+    new_code = Vector{Any}(undef, nstmts)
+    ssavalmap = fill(SSAValue(-1), length(ci.ssavaluetypes) + 1)
+    result_types = Any[Any for _ in 1:nstmts]
     # Detect statement positions for assignments and construct array
     for (bb, idx) in bbidxiter(ir)
         stmt = code[idx]
@@ -797,21 +813,21 @@ function construct_ssa!(ci::CodeInfo, code::Vector{Any}, ir::IRCode, domtree::Do
             new_dest = block_for_inst(cfg, stmt.dest)
             if new_dest == bb+1
                 # Drop this node - it's a noop
-                new_code[idx] = stmt.cond
+                new_code[idx] = Expr(:call, GlobalRef(Core, :typeassert), stmt.cond, GlobalRef(Core, :Bool))
             else
                 new_code[idx] = GotoIfNot(stmt.cond, new_dest)
             end
         elseif isexpr(stmt, :enter)
             new_code[idx] = Expr(:enter, block_for_inst(cfg, stmt.args[1]))
             ssavalmap[idx] = SSAValue(idx) # Slot to store token for pop_exception
-        elseif isexpr(stmt, :leave) || isexpr(stmt, :(=)) || isexpr(stmt, :return) ||
+        elseif isexpr(stmt, :leave) || isexpr(stmt, :(=)) || isa(stmt, ReturnNode) ||
             isexpr(stmt, :meta) || isa(stmt, NewvarNode)
             new_code[idx] = stmt
         else
             ssavalmap[idx] = SSAValue(idx)
             result_types[idx] = ci.ssavaluetypes[idx]
             if isa(stmt, PhiNode)
-                edges = Any[edge == 0 ? 0 : block_for_inst(cfg, edge) for edge in stmt.edges]
+                edges = Int32[edge == 0 ? 0 : block_for_inst(cfg, Int(edge)) for edge in stmt.edges]
                 new_code[idx] = PhiNode(edges, stmt.values)
             else
                 new_code[idx] = stmt
@@ -824,16 +840,17 @@ function construct_ssa!(ci::CodeInfo, code::Vector{Any}, ir::IRCode, domtree::Do
             # TODO: This could just be the ones that depend on other phis
             push!(type_refine_phi, ssa.id)
             new_idx = ssa.id
-            node = ir.new_nodes[new_idx]
-            for i = 1:length(node.node.values)
-                orig_typ = typ = typ_for_val(node.node.values[i], ci, sptypes, -1, slottypes)
+            node = new_nodes.stmts[new_idx]
+            phic_values = (node[:inst]::PhiCNode).values
+            for i = 1:length(phic_values)
+                orig_typ = typ = typ_for_val(phic_values[i], ci, ir.sptypes, -1, slottypes)
                 @assert !isa(typ, MaybeUndef)
                 while isa(typ, DelayedTyp)
                     typ = types(ir)[typ.phi::NewSSAValue]
                 end
                 new_typ = tmerge(new_typ, typ)
             end
-            ir.new_nodes[new_idx] = NewNode(node.pos, node.attach_after, new_typ, node.node, node.line)
+            node[:type] = new_typ
         end
     end
     # This is a bit awkward, because it basically duplicates what type
@@ -843,27 +860,39 @@ function construct_ssa!(ci::CodeInfo, code::Vector{Any}, ir::IRCode, domtree::Do
     while changed
         changed = false
         for new_idx in type_refine_phi
-            node = ir.new_nodes[new_idx]
-            new_typ = recompute_type(node.node, ci, ir, sptypes, slottypes)
-            if !(node.typ ⊑ new_typ) || !(new_typ ⊑ node.typ)
-                ir.new_nodes[new_idx] = NewNode(node.pos, node.attach_after, new_typ, node.node, node.line)
+            node = new_nodes.stmts[new_idx]
+            new_typ = recompute_type(node[:inst]::Union{PhiNode,PhiCNode}, ci, ir, ir.sptypes, slottypes)
+            if !(node[:type] ⊑ new_typ) || !(new_typ ⊑ node[:type])
+                node[:type] = new_typ
                 changed = true
             end
         end
     end
-    result_types = Any[isa(result_types[i], DelayedTyp) ? types(ir)[result_types[i].phi::NewSSAValue] : result_types[i] for i in 1:length(result_types)]
-    new_nodes = NewNode[let node = ir.new_nodes[i]
-            typ = isa(node.typ, DelayedTyp) ? types(ir)[node.typ.phi::NewSSAValue] : node.typ
-            NewNode(node.pos, node.attach_after, typ, node.node, node.line)
-        end for i in 1:length(ir.new_nodes)]
+    for i in 1:length(result_types)
+        rt_i = result_types[i]
+        if rt_i isa DelayedTyp
+            result_types[i] = types(ir)[rt_i.phi::NewSSAValue]
+        end
+    end
+    for i = 1:length(new_nodes)
+        local node = new_nodes.stmts[i]
+        local typ = node[:type]
+        if isa(typ, DelayedTyp)
+            node[:type] = types(ir)[typ.phi::NewSSAValue]
+        end
+    end
     # Renumber SSA values
-    new_code = Any[new_to_regular(renumber_ssa!(new_code[i], ssavalmap), length(ir.stmts)) for i in 1:length(new_code)]
-    new_nodes = NewNode[let node = new_nodes[i]
-            NewNode(node.pos, node.attach_after, node.typ,
-            new_to_regular(renumber_ssa!(node.node, ssavalmap), length(ir.stmts)),
-            node.line)
-        end for i in 1:length(new_nodes)]
-    ir = IRCode(ir, new_code, result_types, ir.lines, ir.flags, ir.cfg, new_nodes)
+    @assert isempty(ir.stmts.type)
+    resize!(ir.stmts.type, nstmts)
+    for i in 1:nstmts
+        local node = ir.stmts[i]
+        node[:inst] = new_to_regular(renumber_ssa!(new_code[i], ssavalmap), nstmts)
+        node[:type] = result_types[i]
+    end
+    for i = 1:length(new_nodes)
+        local node = new_nodes.stmts[i]
+        node[:inst] = new_to_regular(renumber_ssa!(node[:inst], ssavalmap), nstmts)
+    end
     @timeit "domsort" ir = domsort_ssa!(ir, domtree)
     return ir
 end
