@@ -91,7 +91,7 @@ JL_DLLEXPORT complex_t *cptest(complex_t *a) {
 
 JL_DLLEXPORT complex_t *cptest_static(complex_t *a) {
     if (verbose) fprintf(stderr,"%" PRIjint " + %" PRIjint " i\n", a->real, a->imag);
-    complex_t *b = (complex_t*)malloc(sizeof(complex_t));
+    complex_t *b = (complex_t*)malloc_s(sizeof(complex_t));
     b->real = a->real;
     b->imag = a->imag;
     return b;
@@ -856,7 +856,9 @@ typedef struct {
 
 JL_DLLEXPORT struct_aa64_3 test_aa64_vec_2(struct_aa64_3 v1, struct_aa64_4 v2)
 {
-    struct_aa64_3 x = {v1.v1 + vmovn_s16(v2.v1), v1.v2 - v2.v2};
+    // The cast below is to workaround GCC issue.
+    // https://gcc.gnu.org/bugzilla/show_bug.cgi?id=96990
+    struct_aa64_3 x = {(int8x8_t)(v1.v1 + vmovn_s16(v2.v1)), (float32x2_t)(v1.v2 - v2.v2)};
     return x;
 }
 
@@ -978,3 +980,6 @@ JL_DLLEXPORT int threadcall_args(int a, int b) {
 JL_DLLEXPORT void c_exit_finalizer(void* v) {
     printf("c_exit_finalizer: %d, %u", *(int*)v, (unsigned)((uintptr_t)v & (uintptr_t)1));
 }
+
+// global variable for cglobal testing
+JL_DLLEXPORT const int global_var = 1;
