@@ -639,13 +639,14 @@ static void cvalue_printdata(fl_context_t *fl_ctx, ios_t *f, void *data,
             // These states should be context independent.
             static size_t (*volatile jl_static_print)(ios_t*, void*) = NULL;
             static volatile int init = 0;
+            // XXX: use uv_once
             if (init == 0) {
 #if defined(RTLD_SELF)
                 jl_static_print = (size_t (*)(ios_t*, void*))
-                    dlsym(RTLD_SELF, "jl_static_show");
+                    (uintptr_t)dlsym(RTLD_SELF, "jl_static_show");
 #elif defined(RTLD_DEFAULT)
                 jl_static_print = (size_t (*)(ios_t*, void*))
-                    dlsym(RTLD_DEFAULT, "jl_static_show");
+                    (uintptr_t)dlsym(RTLD_DEFAULT, "jl_static_show");
 #elif defined(_OS_WINDOWS_)
                 HMODULE handle;
                 if (GetModuleHandleExW(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS |
@@ -653,7 +654,7 @@ static void cvalue_printdata(fl_context_t *fl_ctx, ios_t *f, void *data,
                                        (LPCWSTR)(&cvalue_printdata),
                                        &handle)) {
                     jl_static_print = (size_t (*)(ios_t*, void*))
-                        GetProcAddress(handle, "jl_static_show");
+                        (uintptr_t)GetProcAddress(handle, "jl_static_show");
                 }
 #endif
                 init = 1;
@@ -748,8 +749,7 @@ static void cvalue_print(fl_context_t *fl_ctx, ios_t *f, value_t v)
         void *fptr = *(void**)data;
         label = (value_t)ptrhash_get(&fl_ctx->reverse_dlsym_lookup_table, cv);
         if (label == (value_t)HT_NOTFOUND) {
-            fl_ctx->HPOS += ios_printf(f, "#<builtin @0x%08zx>",
-                                  (size_t)(builtin_t)fptr);
+            fl_ctx->HPOS += ios_printf(f, "#<builtin @0x%08zx>", (size_t)fptr);
         }
         else {
             if (fl_ctx->print_princ) {
