@@ -23,7 +23,7 @@
 #define MIN_STACK_MAPPINGS_PER_POOL 5
 
 const size_t jl_guard_size = (4096 * 8);
-static uint32_t num_stack_mappings = 0;
+static _Atomic(uint32_t) num_stack_mappings = 0;
 
 #ifdef _OS_WINDOWS_
 #define MAP_FAILED NULL
@@ -119,7 +119,8 @@ static void _jl_free_stack(jl_ptls_t ptls, void *stkbuf, size_t bufsz)
 
 JL_DLLEXPORT void jl_free_stack(void *stkbuf, size_t bufsz)
 {
-    _jl_free_stack(jl_get_ptls_states(), stkbuf, bufsz);
+    jl_task_t *ct = jl_current_task;
+    _jl_free_stack(ct->ptls, stkbuf, bufsz);
 }
 
 
@@ -142,7 +143,8 @@ void jl_release_task_stack(jl_ptls_t ptls, jl_task_t *task)
 
 JL_DLLEXPORT void *jl_malloc_stack(size_t *bufsz, jl_task_t *owner) JL_NOTSAFEPOINT
 {
-    jl_ptls_t ptls = jl_get_ptls_states();
+    jl_task_t *ct = jl_current_task;
+    jl_ptls_t ptls = ct->ptls;
     size_t ssize = *bufsz;
     void *stk = NULL;
     if (ssize <= pool_sizes[JL_N_STACK_POOLS - 1]) {
@@ -216,8 +218,12 @@ void sweep_stack_pools(void)
             continue;
         while (1) {
             jl_task_t *t = (jl_task_t*)lst[n];
+            assert(jl_is_task(t));
             if (gc_marked(jl_astaggedvalue(t)->bits.gc)) {
-                n++;
+                if (t->stkbuf == NULL)
+                    ndel++; // jl_release_task_stack called
+                else
+                    n++;
             }
             else {
                 ndel++;
@@ -243,3 +249,31 @@ void sweep_stack_pools(void)
         live_tasks->len -= ndel;
     }
 }
+
+JL_DLLEXPORT jl_array_t *jl_live_tasks(void)
+{
+    jl_task_t *ct = jl_current_task;
+    jl_ptls_t ptls = ct->ptls;
+    arraylist_t *live_tasks = &ptls->heap.live_tasks;
+    size_t i, j, l;
+    jl_array_t *a;
+    do {
+        l = live_tasks->len;
+        a = jl_alloc_vec_any(l + 1); // may gc, changing the number of tasks
+    } while (l + 1 < live_tasks->len);
+    l = live_tasks->len;
+    void **lst = live_tasks->items;
+    j = 0;
+    ((void**)jl_array_data(a))[j++] = ptls->root_task;
+    for (i = 0; i < l; i++) {
+        if (((jl_task_t*)lst[i])->stkbuf != NULL)
+            ((void**)jl_array_data(a))[j++] = lst[i];
+    }
+    l = jl_array_len(a);
+    if (j < l) {
+        JL_GC_PUSH1(&a);
+        jl_array_del_end(a, l - j);
+        JL_GC_POP();
+    }
+    return a;
+}
