@@ -57,7 +57,7 @@ void jl_init_uv(void)
     JL_MUTEX_INIT(&jl_uv_mutex); // a file-scope initializer can be used instead
 }
 
-int jl_uv_n_waiters = 0;
+_Atomic(int) jl_uv_n_waiters = 0;
 
 void JL_UV_LOCK(void)
 {
@@ -105,11 +105,11 @@ static void jl_uv_closeHandle(uv_handle_t *handle)
         JL_STDERR = (JL_STREAM*)STDERR_FILENO;
     // also let the client app do its own cleanup
     if (handle->type != UV_FILE && handle->data) {
-        jl_ptls_t ptls = jl_get_ptls_states();
-        size_t last_age = ptls->world_age;
-        ptls->world_age = jl_world_counter;
+        jl_task_t *ct = jl_current_task;
+        size_t last_age = ct->world_age;
+        ct->world_age = jl_world_counter;
         jl_uv_call_close_callback((jl_value_t*)handle->data);
-        ptls->world_age = last_age;
+        ct->world_age = last_age;
     }
     if (handle == (uv_handle_t*)&signal_async)
         return;
@@ -205,16 +205,17 @@ extern volatile unsigned _threadedregion;
 
 JL_DLLEXPORT int jl_process_events(void)
 {
-    jl_ptls_t ptls = jl_get_ptls_states();
+    jl_task_t *ct = jl_current_task;
     uv_loop_t *loop = jl_io_loop;
-    if (loop && (_threadedregion || ptls->tid == 0)) {
-        jl_gc_safepoint_(ptls);
+    jl_gc_safepoint_(ct->ptls);
+    if (loop && (_threadedregion || ct->tid == 0)) {
         if (jl_atomic_load(&jl_uv_n_waiters) == 0 && jl_mutex_trylock(&jl_uv_mutex)) {
             loop->stop_flag = 0;
             int r = uv_run(loop, UV_RUN_NOWAIT);
             JL_UV_UNLOCK();
             return r;
         }
+        jl_gc_safepoint_(ct->ptls);
     }
     return 0;
 }
@@ -403,9 +404,9 @@ JL_DLLEXPORT int jl_fs_access(char *path, int mode)
 JL_DLLEXPORT int jl_fs_write(uv_os_fd_t handle, const char *data, size_t len,
                              int64_t offset) JL_NOTSAFEPOINT
 {
-    jl_ptls_t ptls = jl_get_ptls_states();
+    jl_task_t *ct = jl_get_current_task();
     // TODO: fix this cheating
-    if (ptls->safe_restore || ptls->tid != 0)
+    if (jl_get_safe_restore() || ct == NULL || ct->tid != 0)
 #ifdef _OS_WINDOWS_
         return WriteFile(handle, data, len, NULL, NULL);
 #else
@@ -504,8 +505,8 @@ JL_DLLEXPORT void jl_uv_puts(uv_stream_t *stream, const char *str, size_t n)
     }
 
     // TODO: Hack to make CoreIO thread-safer
-    jl_ptls_t ptls = jl_get_ptls_states();
-    if (ptls->tid != 0) {
+    jl_task_t *ct = jl_get_current_task();
+    if (ct == NULL || ct->tid != 0) {
         if (stream == JL_STDOUT) {
             fd = UV_STDOUT_FD;
         }
@@ -861,7 +862,7 @@ JL_DLLEXPORT int jl_tcp_quickack(uv_tcp_t *handle, int on)
 
 JL_DLLEXPORT int jl_has_so_reuseport(void)
 {
-#if defined(SO_REUSEPORT)
+#if defined(SO_REUSEPORT) && !defined(_OS_DARWIN_)
     return 1;
 #else
     return 0;
