@@ -40,11 +40,11 @@ uint64_t io_wakeup_leave;
 JL_DLLEXPORT int jl_set_task_tid(jl_task_t *task, int tid) JL_NOTSAFEPOINT
 {
     // Try to acquire the lock on this task.
-    int16_t was = task->tid;
+    int16_t was = jl_atomic_load_relaxed(&task->tid);
     if (was == tid)
         return 1;
     if (was == -1)
-        return jl_atomic_bool_compare_exchange(&task->tid, -1, tid);
+        return jl_atomic_cmpswap(&task->tid, &was, tid);
     return 0;
 }
 
@@ -57,10 +57,10 @@ extern int jl_gc_mark_queue_obj_explicit(jl_gc_mark_cache_t *gc_cache,
 
 /* a task heap */
 typedef struct taskheap_tag {
-    jl_mutex_t lock;
+    uv_mutex_t lock;
     jl_task_t **tasks;
-    int32_t ntasks;
-    int16_t prio;
+    _Atomic(int32_t) ntasks;
+    _Atomic(int16_t) prio;
 } taskheap_t;
 
 /* multiqueue parameters */
@@ -83,10 +83,10 @@ static inline void multiq_init(void)
     heap_p = heap_c * jl_n_threads;
     heaps = (taskheap_t *)calloc(heap_p, sizeof(taskheap_t));
     for (int32_t i = 0; i < heap_p; ++i) {
-        jl_mutex_init(&heaps[i].lock);
+        uv_mutex_init(&heaps[i].lock);
         heaps[i].tasks = (jl_task_t **)calloc(tasks_per_heap, sizeof(jl_task_t*));
-        heaps[i].ntasks = 0;
-        heaps[i].prio = INT16_MAX;
+        jl_atomic_store_relaxed(&heaps[i].ntasks, 0);
+        jl_atomic_store_relaxed(&heaps[i].prio, INT16_MAX);
     }
     unbias_cong(heap_p, &cong_unbias);
 }
@@ -108,12 +108,12 @@ static inline void sift_up(taskheap_t *heap, int32_t idx)
 
 static inline void sift_down(taskheap_t *heap, int32_t idx)
 {
-    if (idx < heap->ntasks) {
+    if (idx < jl_atomic_load_relaxed(&heap->ntasks)) {
         for (int32_t child = heap_d*idx + 1;
                 child < tasks_per_heap && child <= heap_d*idx + heap_d;
                 ++child) {
             if (heap->tasks[child]
-                    &&  heap->tasks[child]->prio < heap->tasks[idx]->prio) {
+                    && heap->tasks[child]->prio < heap->tasks[idx]->prio) {
                 jl_task_t *t = heap->tasks[idx];
                 heap->tasks[idx] = heap->tasks[child];
                 heap->tasks[child] = t;
@@ -126,26 +126,28 @@ static inline void sift_down(taskheap_t *heap, int32_t idx)
 
 static inline int multiq_insert(jl_task_t *task, int16_t priority)
 {
-    jl_ptls_t ptls = jl_get_ptls_states();
+    jl_ptls_t ptls = jl_current_task->ptls;
     uint64_t rn;
 
     task->prio = priority;
     do {
         rn = cong(heap_p, cong_unbias, &ptls->rngseed);
-    } while (!jl_mutex_trylock_nogc(&heaps[rn].lock));
+    } while (uv_mutex_trylock(&heaps[rn].lock) != 0);
 
-    if (heaps[rn].ntasks >= tasks_per_heap) {
-        jl_mutex_unlock_nogc(&heaps[rn].lock);
+    if (jl_atomic_load_relaxed(&heaps[rn].ntasks) >= tasks_per_heap) {
+        uv_mutex_unlock(&heaps[rn].lock);
         // multiq insertion failed, increase #tasks per heap
         return -1;
     }
 
-    heaps[rn].tasks[heaps[rn].ntasks++] = task;
-    sift_up(&heaps[rn], heaps[rn].ntasks-1);
-    int16_t prio = jl_atomic_load(&heaps[rn].prio);
+    int32_t ntasks = jl_atomic_load_relaxed(&heaps[rn].ntasks);
+    jl_atomic_store_relaxed(&heaps[rn].ntasks, ntasks + 1);
+    heaps[rn].tasks[ntasks] = task;
+    sift_up(&heaps[rn], ntasks);
+    int16_t prio = jl_atomic_load_relaxed(&heaps[rn].prio);
     if (task->prio < prio)
-        jl_atomic_store(&heaps[rn].prio, task->prio);
-    jl_mutex_unlock_nogc(&heaps[rn].lock);
+        jl_atomic_store_relaxed(&heaps[rn].prio, task->prio);
+    uv_mutex_unlock(&heaps[rn].lock);
 
     return 0;
 }
@@ -153,27 +155,28 @@ static inline int multiq_insert(jl_task_t *task, int16_t priority)
 
 static inline jl_task_t *multiq_deletemin(void)
 {
-    jl_ptls_t ptls = jl_get_ptls_states();
+    jl_ptls_t ptls = jl_current_task->ptls;
     uint64_t rn1 = 0, rn2;
     int32_t i;
     int16_t prio1, prio2;
     jl_task_t *task;
  retry:
+    jl_gc_safepoint();
     for (i = 0; i < heap_p; ++i) {
         rn1 = cong(heap_p, cong_unbias, &ptls->rngseed);
         rn2 = cong(heap_p, cong_unbias, &ptls->rngseed);
-        prio1 = jl_atomic_load(&heaps[rn1].prio);
-        prio2 = jl_atomic_load(&heaps[rn2].prio);
+        prio1 = jl_atomic_load_relaxed(&heaps[rn1].prio);
+        prio2 = jl_atomic_load_relaxed(&heaps[rn2].prio);
         if (prio1 > prio2) {
             prio1 = prio2;
             rn1 = rn2;
         }
         else if (prio1 == prio2 && prio1 == INT16_MAX)
             continue;
-        if (jl_mutex_trylock_nogc(&heaps[rn1].lock)) {
-            if (prio1 == heaps[rn1].prio)
+        if (uv_mutex_trylock(&heaps[rn1].lock) == 0) {
+            if (prio1 == jl_atomic_load_relaxed(&heaps[rn1].prio))
                 break;
-            jl_mutex_unlock_nogc(&heaps[rn1].lock);
+            uv_mutex_unlock(&heaps[rn1].lock);
         }
     }
     if (i == heap_p)
@@ -181,18 +184,20 @@ static inline jl_task_t *multiq_deletemin(void)
 
     task = heaps[rn1].tasks[0];
     if (!jl_set_task_tid(task, ptls->tid)) {
-        jl_mutex_unlock_nogc(&heaps[rn1].lock);
+        uv_mutex_unlock(&heaps[rn1].lock);
         goto retry;
     }
-    heaps[rn1].tasks[0] = heaps[rn1].tasks[--heaps[rn1].ntasks];
-    heaps[rn1].tasks[heaps[rn1].ntasks] = NULL;
+    int32_t ntasks = jl_atomic_load_relaxed(&heaps[rn1].ntasks) - 1;
+    jl_atomic_store_relaxed(&heaps[rn1].ntasks, ntasks);
+    heaps[rn1].tasks[0] = heaps[rn1].tasks[ntasks];
+    heaps[rn1].tasks[ntasks] = NULL;
     prio1 = INT16_MAX;
-    if (heaps[rn1].ntasks > 0) {
+    if (ntasks > 0) {
         sift_down(&heaps[rn1], 0);
         prio1 = heaps[rn1].tasks[0]->prio;
     }
-    jl_atomic_store(&heaps[rn1].prio, prio1);
-    jl_mutex_unlock_nogc(&heaps[rn1].lock);
+    jl_atomic_store_relaxed(&heaps[rn1].prio, prio1);
+    uv_mutex_unlock(&heaps[rn1].lock);
 
     return task;
 }
@@ -202,7 +207,7 @@ void jl_gc_mark_enqueued_tasks(jl_gc_mark_cache_t *gc_cache, jl_gc_mark_sp_t *sp
 {
     int32_t i, j;
     for (i = 0; i < heap_p; ++i)
-        for (j = 0; j < heaps[i].ntasks; ++j)
+        for (j = 0; j < jl_atomic_load_relaxed(&heaps[i].ntasks); ++j)
             jl_gc_mark_queue_obj_explicit(gc_cache, sp, (jl_value_t *)heaps[i].tasks[j]);
 }
 
@@ -211,7 +216,7 @@ static int multiq_check_empty(void)
 {
     int32_t i;
     for (i = 0; i < heap_p; ++i) {
-        if (heaps[i].ntasks != 0)
+        if (jl_atomic_load_relaxed(&heaps[i].ntasks) != 0)
             return 0;
     }
     return 1;
@@ -223,12 +228,14 @@ static int multiq_check_empty(void)
 // ---
 
 // initialize the threading infrastructure
+// (used only by the main thread)
 void jl_init_threadinginfra(void)
 {
     /* initialize the synchronization trees pool and the multiqueue */
     multiq_init();
 
-    jl_ptls_t ptls = jl_get_ptls_states();
+    jl_ptls_t ptls = jl_current_task->ptls;
+    jl_install_thread_signal_handler(ptls);
     uv_mutex_init(&ptls->sleep_lock);
     uv_cond_init(&ptls->wake_signal);
 }
@@ -242,12 +249,13 @@ void jl_threadfun(void *arg)
     jl_threadarg_t *targ = (jl_threadarg_t*)arg;
 
     // initialize this thread (set tid, create heap, set up root task)
-    jl_init_threadtls(targ->tid);
+    jl_ptls_t ptls = jl_init_threadtls(targ->tid);
     void *stack_lo, *stack_hi;
     jl_init_stack_limits(0, &stack_lo, &stack_hi);
-    jl_init_root_task(stack_lo, stack_hi);
-
-    jl_ptls_t ptls = jl_get_ptls_states();
+    // warning: this changes `jl_current_task`, so be careful not to call that from this function
+    jl_task_t *ct = jl_init_root_task(ptls, stack_lo, stack_hi);
+    JL_GC_PROMISE_ROOTED(ct);
+    jl_install_thread_signal_handler(ptls);
 
     // set up sleep mechanism for this thread
     uv_mutex_init(&ptls->sleep_lock);
@@ -261,7 +269,7 @@ void jl_threadfun(void *arg)
     free(targ);
 
     (void)jl_gc_unsafe_enter(ptls);
-    jl_finish_task(jl_current_task); // noreturn
+    jl_finish_task(ct); // noreturn
 }
 
 
@@ -279,22 +287,20 @@ int jl_running_under_rr(int recheck)
 #ifdef _OS_LINUX_
 #define RR_CALL_BASE 1000
 #define SYS_rrcall_check_presence (RR_CALL_BASE + 8)
-    static int checked_running_under_rr = 0;
-    static int is_running_under_rr = 0;
-    if (!checked_running_under_rr || recheck) {
+    static _Atomic(int) is_running_under_rr = 0;
+    int rr = jl_atomic_load_relaxed(&is_running_under_rr);
+    if (rr == 0 || recheck) {
         int ret = syscall(SYS_rrcall_check_presence, 0, 0, 0, 0, 0, 0);
-        if (ret == -1) {
+        if (ret == -1)
             // Should always be ENOSYS, but who knows what people do for
             // unknown syscalls with their seccomp filters, so just say
             // that we don't have rr.
-            is_running_under_rr = 0;
-        }
-        else {
-            is_running_under_rr = 1;
-        }
-        checked_running_under_rr = 1;
+            rr = 2;
+        else
+            rr = 1;
+        jl_atomic_store_relaxed(&is_running_under_rr, rr);
     }
-    return is_running_under_rr;
+    return rr == 1;
 #else
     return 0;
 #endif
@@ -329,7 +335,9 @@ static int sleep_check_after_threshold(uint64_t *start_cycles)
 static void wake_thread(int16_t tid)
 {
     jl_ptls_t other = jl_all_tls_states[tid];
-    if (jl_atomic_bool_compare_exchange(&other->sleep_check_state, sleeping, not_sleeping)) {
+    int8_t state = sleeping;
+    jl_atomic_cmpswap(&other->sleep_check_state, &state, not_sleeping);
+    if (state == sleeping) {
         uv_mutex_lock(&other->sleep_lock);
         uv_cond_signal(&other->wake_signal);
         uv_mutex_unlock(&other->sleep_lock);
@@ -347,24 +355,24 @@ static void wake_libuv(void)
 /* ensure thread tid is awake if necessary */
 JL_DLLEXPORT void jl_wakeup_thread(int16_t tid)
 {
-    jl_ptls_t ptls = jl_get_ptls_states();
-    jl_thread_t uvlock = jl_atomic_load(&jl_uv_mutex.owner);
-    int16_t self = ptls->tid;
-    jl_thread_t system_self = jl_all_tls_states[self]->system_id;
+    jl_task_t *ct = jl_current_task;
+    jl_ptls_t ptls = ct->ptls;
+    jl_task_t *uvlock = jl_atomic_load(&jl_uv_mutex.owner);
+    int16_t self = jl_atomic_load_relaxed(&ct->tid);
     JULIA_DEBUG_SLEEPWAKE( wakeup_enter = cycleclock() );
     if (tid == self || tid == -1) {
         // we're already awake, but make sure we'll exit uv_run
         if (jl_atomic_load_relaxed(&ptls->sleep_check_state) == sleeping)
             jl_atomic_store(&ptls->sleep_check_state, not_sleeping);
-        if (uvlock == system_self)
+        if (uvlock == ct)
             uv_stop(jl_global_event_loop());
     }
     else {
         // something added to the sticky-queue: notify that thread
         wake_thread(tid);
         // check if we need to notify uv_run too
-        jl_thread_t system_tid = jl_all_tls_states[tid]->system_id;
-        if (uvlock != system_self && jl_atomic_load(&jl_uv_mutex.owner) == system_tid)
+        jl_task_t *system_tid = jl_atomic_load_relaxed(&jl_all_tls_states[tid]->current_task);
+        if (uvlock != ct && jl_atomic_load(&jl_uv_mutex.owner) == system_tid)
             wake_libuv();
     }
     // check if the other threads might be sleeping
@@ -377,7 +385,7 @@ JL_DLLEXPORT void jl_wakeup_thread(int16_t tid)
                 wake_thread(tid);
         }
         // check if we need to notify uv_run too
-        if (uvlock != system_self && jl_atomic_load(&jl_uv_mutex.owner) != 0)
+        if (uvlock != ct && jl_atomic_load(&jl_uv_mutex.owner) != NULL)
             wake_libuv();
     }
     JULIA_DEBUG_SLEEPWAKE( wakeup_leave = cycleclock() );
@@ -391,15 +399,14 @@ static jl_task_t *get_next_task(jl_value_t *trypoptask, jl_value_t *q)
     jl_value_t *args[2] = { trypoptask, q };
     jl_task_t *task = (jl_task_t*)jl_apply(args, 2);
     if (jl_typeis(task, jl_task_type)) {
-        int self = jl_get_ptls_states()->tid;
+        int self = jl_atomic_load_relaxed(&jl_current_task->tid);
         jl_set_task_tid(task, self);
         return task;
     }
-    jl_gc_safepoint();
     return multiq_deletemin();
 }
 
-static int may_sleep(jl_ptls_t ptls)
+static int may_sleep(jl_ptls_t ptls) JL_NOTSAFEPOINT
 {
     // sleep_check_state is only transitioned from not_sleeping to sleeping
     // by the thread itself. As a result, if this returns false, it will
@@ -407,16 +414,15 @@ static int may_sleep(jl_ptls_t ptls)
     return jl_atomic_load_relaxed(&ptls->sleep_check_state) == sleeping;
 }
 
-extern volatile unsigned _threadedregion;
+extern _Atomic(unsigned) _threadedregion;
 
 JL_DLLEXPORT jl_task_t *jl_task_get_next(jl_value_t *trypoptask, jl_value_t *q)
 {
-    jl_ptls_t ptls = jl_get_ptls_states();
+    jl_task_t *ct = jl_current_task;
     uint64_t start_cycles = 0;
-    jl_task_t *task;
 
     while (1) {
-        task = get_next_task(trypoptask, q);
+        jl_task_t *task = get_next_task(trypoptask, q);
         if (task)
             return task;
 
@@ -428,14 +434,17 @@ JL_DLLEXPORT jl_task_t *jl_task_get_next(jl_value_t *trypoptask, jl_value_t *q)
         }
 
         jl_cpu_pause();
-        if (sleep_check_after_threshold(&start_cycles) || (!_threadedregion && ptls->tid == 0)) {
+        jl_ptls_t ptls = ct->ptls;
+        if (sleep_check_after_threshold(&start_cycles) || (!jl_atomic_load_relaxed(&_threadedregion) && ptls->tid == 0)) {
             jl_atomic_store(&ptls->sleep_check_state, sleeping); // acquire sleep-check lock
             if (!multiq_check_empty()) {
                 if (jl_atomic_load_relaxed(&ptls->sleep_check_state) != not_sleeping)
                     jl_atomic_store(&ptls->sleep_check_state, not_sleeping); // let other threads know they don't need to wake us
                 continue;
             }
-            task = get_next_task(trypoptask, q);
+            task = get_next_task(trypoptask, q); // WARNING: this should not yield
+            if (ptls != ct->ptls)
+                continue; // oops, get_next_task did yield--start over
             if (task) {
                 if (jl_atomic_load_relaxed(&ptls->sleep_check_state) != not_sleeping)
                     jl_atomic_store(&ptls->sleep_check_state, not_sleeping); // let other threads know they don't need to wake us
@@ -448,7 +457,7 @@ JL_DLLEXPORT jl_task_t *jl_task_get_next(jl_value_t *trypoptask, jl_value_t *q)
             // outside of threaded regions, all IO is permitted,
             // but only on thread 1
             int uvlock = 0;
-            if (_threadedregion) {
+            if (jl_atomic_load_relaxed(&_threadedregion)) {
                 uvlock = jl_mutex_trylock(&jl_uv_mutex);
             }
             else if (ptls->tid == 0) {
@@ -475,7 +484,7 @@ JL_DLLEXPORT jl_task_t *jl_task_get_next(jl_value_t *trypoptask, jl_value_t *q)
                     JL_UV_UNLOCK();
                     // optimization: check again first if we may have work to do
                     if (!may_sleep(ptls)) {
-                        assert(ptls->sleep_check_state == not_sleeping);
+                        assert(jl_atomic_load_relaxed(&ptls->sleep_check_state) == not_sleeping);
                         start_cycles = 0;
                         continue;
                     }
@@ -490,7 +499,7 @@ JL_DLLEXPORT jl_task_t *jl_task_get_next(jl_value_t *trypoptask, jl_value_t *q)
                         continue;
                     }
                 }
-                if (!_threadedregion && active && ptls->tid == 0) {
+                if (!jl_atomic_load_relaxed(&_threadedregion) && active && ptls->tid == 0) {
                     // thread 0 is the only thread permitted to run the event loop
                     // so it needs to stay alive
                     if (jl_atomic_load_relaxed(&ptls->sleep_check_state) != not_sleeping)
@@ -508,7 +517,7 @@ JL_DLLEXPORT jl_task_t *jl_task_get_next(jl_value_t *trypoptask, jl_value_t *q)
                 uv_cond_wait(&ptls->wake_signal, &ptls->sleep_lock);
                 // TODO: help with gc work here, if applicable
             }
-            assert(ptls->sleep_check_state == not_sleeping);
+            assert(jl_atomic_load_relaxed(&ptls->sleep_check_state) == not_sleeping);
             uv_mutex_unlock(&ptls->sleep_lock);
             JULIA_DEBUG_SLEEPWAKE( ptls->sleep_leave = cycleclock() );
             jl_gc_safe_leave(ptls, gc_state); // contains jl_gc_safepoint
