@@ -12,7 +12,7 @@ using Test, LinearAlgebra
     lq,
     lu,
     qr,
-    x -> qr(x, Val(true)),
+    x -> qr(x, ColumnNorm()),
     svd,
     schur,
 ]
@@ -37,13 +37,8 @@ using Test, LinearAlgebra
         return x isa AbstractArray{Float64} ? Float64.(Float32.(x)) : x
     end...)
 
-    if f === qr
-        @test F == G
-        @test isequal(F, G)
-    else
-        @test_broken F == G
-        @test_broken isequal(F, G)
-    end
+    @test F == G broken=!(f === eigen || f === qr)
+    @test isequal(F, G) broken=!(f === eigen || f === qr)
     @test hash(F) == hash(G)
 end
 
