@@ -726,4 +726,13 @@ end
     @test extrema(x; dims=[]) == extrema(y; dims=[])
 end
 
+@testset "issue #42670 - error in sparsevec outer product" begin
+    A = spzeros(Int, 4)
+    B = copy(A)
+    C = sparsevec([0 0 1 1 0 0])'
+    A[2] = 1
+    A[2] = 0
+    @test A * C == B * C == spzeros(Int, 4, 6)
+end
+
 end # module
