@@ -18,6 +18,9 @@ if !@isdefined(testenv_defined)
         push!(test_exeflags.exec, "--startup-file=no")
         push!(test_exeflags.exec, "--depwarn=error")
     end
+    if haskey(ENV, "JULIA_TEST_EXTRA_EXEFLAGS")
+        append!(test_exeflags.exec, Base.shell_split(ENV["JULIA_TEST_EXTRA_EXEFLAGS"]))
+    end
 
     if haskey(ENV, "JULIA_TEST_EXENAME")
         popfirst!(test_exeflags.exec)
